from typing import Generator
from typing import List
from langchain.memory import ConversationBufferMemory
from langchain.prompts import ChatPromptTemplate, MessagesPlaceholder, SystemMessagePromptTemplate, HumanMessagePromptTemplate
from langchain.schema import AIMessage, HumanMessage
from langchain.text_splitter import CharacterTextSplitter
from langchain_community.vectorstores import FAISS
from sonika_langchain_bot.langchain_bdi import Belief, BotBeliefSystem
from sonika_langchain_bot.langchain_class import FileProcessorInterface, IEmbeddings, ILanguageModel, Message, ResponseModel
from langgraph.checkpoint.memory import MemorySaver
from langgraph.prebuilt import create_react_agent
from langchain_community.tools import BaseTool
import re

class LangChainBot:
    """
    Implementación principal del bot conversacional con capacidades de procesamiento de archivos,
    memoria de conversación y uso de herramientas personalizadas.
    """

    def __init__(self, language_model: ILanguageModel, embeddings: IEmbeddings, beliefs: List[Belief], tools: List[BaseTool]):
        """
        Inicializa el bot con el modelo de lenguaje, embeddings y herramientas necesarias.

        Args:
            language_model (ILanguageModel): Modelo de lenguaje a utilizar
            embeddings (IEmbeddings): Modelo de embeddings para procesamiento de texto
            beliefs (List[Belief]): Lista de creencias iniciales
            tools (List[BaseTool]): Lista de herramientas disponibles
        """
        self.language_model = language_model
        self.embeddings = embeddings
        self.memory = ConversationBufferMemory(return_messages=True)
        self.memory_agent = MemorySaver()
        self.vector_store = None
        self.tools = tools
        self.beliefs = beliefs
        self.belief_system = BotBeliefSystem('Hal9000', '1.0.0', beliefs_init=beliefs, tools=tools)
        self.conversation = self._create_conversation_chain()
        self.agent_executor = self._create_agent_executor()

    def _create_conversation_chain(self):
        """
        Crea la cadena de conversación con el prompt template y la memoria.
        Ahora incluye el contexto del sistema de creencias.
        """
        beliefs_context = self.belief_system.generate_prompt_context()
        full_system_prompt = f"{beliefs_context}\n\n"

        prompt = ChatPromptTemplate.from_messages([
            SystemMessagePromptTemplate.from_template(full_system_prompt),
            MessagesPlaceholder(variable_name="history"),
            HumanMessagePromptTemplate.from_template("{input}")
        ])

        # Usando RunnableSequence para reemplazar LLMChain
        return prompt | self.language_model.model 

    def _create_agent_executor(self):
        """
        Crea el ejecutor del agente con las herramientas configuradas.

        Returns:
            Agent: Agente configurado con las herramientas
        """
        return create_react_agent(self.language_model.model, self.tools, checkpointer=self.memory_agent)

    def _getInstruccionTool(self, bot_response):
        """
        Extrae las instrucciones para herramientas del texto de respuesta del bot.

        Args:
            bot_response (str): Respuesta del bot a analizar

        Returns:
            str: Instrucción extraída o cadena vacía si no se encuentra
        """
        patron = r'\*\*\*(.*?)\*\*\*'
        coincidencia = re.search(patron, bot_response)
        return coincidencia.group(1).strip() if coincidencia else ''

    def get_response(self, user_input: str) -> ResponseModel:
        """
        Genera una respuesta para la entrada del usuario, procesando el contexto y ejecutando herramientas si es necesario.

        Args:
            user_input (str): Entrada del usuario

        Returns:
            ResponseModel: Modelo de respuesta con tokens y texto
        """
        context = self._get_context(user_input)
        augmented_input = f"User question: {user_input}"
        if context:
            augmented_input = f"Context from attached files:\n{context}\n\nUser question: {user_input}"

        bot_response = self.conversation.invoke({"input": augmented_input, "history": self.memory.chat_memory.messages})

        token_usage= bot_response.response_metadata['token_usage']
        bot_response = bot_response.content
        
        instruction_tool = self._getInstruccionTool(bot_response)

        if instruction_tool:
            messages = [HumanMessage(content=instruction_tool)]
            thread_id = "abc123"
            config = {"configurable": {"thread_id": thread_id}}

            result_stream = self.agent_executor.stream(
                {"messages": messages}, config
            )

            tool_response = ""
            agent_response = ""

            for response in result_stream:
                if 'tools' in response:
                    for message in response['tools']['messages']:
                        tool_response = message.content
                if 'agent' in response:
                    for message in response['agent']['messages']:
                        agent_response = message.content

            bot_response = agent_response if agent_response else tool_response

        user_tokens = token_usage['prompt_tokens']
        bot_tokens = token_usage['completion_tokens']

        self.save_messages(user_input, bot_response)

        return ResponseModel(user_tokens=user_tokens, bot_tokens=bot_tokens, response=bot_response)
    
    def get_response_stream(self, user_input: str) -> Generator[str, None, None]:
        """
        Genera una respuesta en streaming para la entrada del usuario, procesando el contexto.

        Args:
            user_input (str): Entrada del usuario

        Yields:
            str: Fragmentos de la respuesta generada por el modelo en tiempo real
        """
        context = self._get_context(user_input)
        augmented_input = f"User question: {user_input}"
        if context:
            augmented_input = f"Context from attached files:\n{context}\n\nUser question: {user_input}"

        result_stream = self.conversation.stream({"input": augmented_input, "history": self.memory.chat_memory.messages})
        for response in result_stream:
            yield response.content


    def _get_context(self, query: str) -> str:
        """
        Obtiene el contexto relevante para una consulta del almacén de vectores.

        Args:
            query (str): Consulta para buscar contexto

        Returns:
            str: Contexto encontrado o cadena vacía
        """
        if self.vector_store:
            docs = self.vector_store.similarity_search(query)
            return "\n".join([doc.page_content for doc in docs])
        return ""

    def clear_memory(self):
        """
        Limpia la memoria de conversación y el almacén de vectores.
        """
        self.memory.clear()
        self.vector_store = None

    def load_conversation_history(self, messages: List[Message]):
        """
        Carga el historial de conversación previo usando la estructura de mensajes simplificada.

        Args:
            messages: Lista de objetos Message que representan cada mensaje.
        """
        for message in messages:
            if message.is_bot:
                self.memory.chat_memory.add_message(AIMessage(content=message.content))
            else:
                self.memory.chat_memory.add_message(HumanMessage(content=message.content))

    def save_messages(self, user_message: str, bot_response: str):
        """
        Guarda los mensajes en la memoria de conversación.

        Args:
            user_message (str): Mensaje del usuario
            bot_response (str): Respuesta del bot
        """
        self.memory.chat_memory.add_message(HumanMessage(content=user_message))
        self.memory.chat_memory.add_message(AIMessage(content=bot_response))

    def process_file(self, file: FileProcessorInterface):
        """
        Procesa un archivo y lo añade al almacén de vectores.

        Args:
            file (FileProcessorInterface): Archivo a procesar
        """
        document = file.getText()
        text_splitter = CharacterTextSplitter(chunk_size=1000, chunk_overlap=0)
        texts = text_splitter.split_documents(document)

        if self.vector_store is None:
            self.vector_store = FAISS.from_texts([doc.page_content for doc in texts], self.embeddings)
        else:
            self.vector_store.add_texts([doc.page_content for doc in texts])



from dataclasses import dataclass, field
from typing import Dict, List, Optional, Set
from datetime import datetime
import json
from enum import Enum
from langchain.tools import Tool
import inspect

class BeliefType(Enum):
    CORE = "core"              # Creencias fundamentales sobre sí mismo
    PERSONALITY = "personality" # Rasgos de personalidad
    USER = "user"              # Creencias sobre el usuario
    CONTEXT = "context"        # Creencias sobre el contexto actual
    KNOWLEDGE = "knowledge"    # Conocimiento general
    TEMPORAL = "temporal"      # Creencias temporales
    TOOLS = "tools"      # Creencias temporales


@dataclass
class Belief:
    """Representa una creencia individual del bot."""
    content: str
    type: BeliefType
    confidence: float = 1.0
    created_at: datetime = field(default_factory=datetime.now)
    last_updated: datetime = field(default_factory=datetime.now)
    source: str = "initial"

@dataclass
class Desire:
    """Representa un deseo o objetivo del bot."""
    description: str
    priority: int
    conditions: List[str]
    achieved: bool = False

@dataclass
class Intention:
    """Representa una intención o plan de acción del bot."""
    description: str
    steps: List[str]
    related_desire: str
    completed: bool = False

class BotBeliefSystem:
    """Sistema principal de creencias del bot."""
    
    def __init__(self, bot_name: str, bot_version: str, tools: List[Tool], beliefs_init: List[Belief]):
        self.bot_name = bot_name
        self.bot_version = bot_version
        self.beliefs_init = beliefs_init
        self.beliefs: Dict[BeliefType, Dict[str, Belief]] = {belief_type: {} for belief_type in BeliefType}
        self.desires: List[Desire] = []
        self.intentions: List[Intention] = []
        self.conversation_history: List[Dict] = []
        self._initialize_core_beliefs()
        self.add_tools_beliefs(tools)
        self.set_beliefs_init()
    
    def set_beliefs_init(self):
        for belief in self.beliefs_init:
            self.add_belief(belief)
        
    def _initialize_core_beliefs(self):
        """Initializes the core beliefs of the bot."""
        core_beliefs = [
            Belief(
                f"My name is {self.bot_name}.",
                BeliefType.CORE,
                confidence=1.0
            ),
            Belief(
                f"I operate on Telegram, and my responses should be as concise and accurate as possible.",
                BeliefType.CORE,
                confidence=1.0
            ),
            # Add more core beliefs as needed
        ]
        
        for belief in core_beliefs:
            self.add_belief(belief)

    def add_tools_beliefs(self, tools: List[Tool]):
        """Carga y procesa las herramientas disponibles para el bot."""
        # Instrucciones sobre el uso de herramientas en inglés
        instrucciones_de_uso = '''\nWhen you want to execute a tool, enclose the command with three asterisks and provide all parameters needed.
                                \nEnsure you gather all relevant information from the conversation to use the parameters.
                                \nIf information is missing, search online.'''
        
        # Agregar un encabezado para la lista de herramientas
        instrucciones_de_uso = instrucciones_de_uso + "\nThis is a list of the tools you can execute:"
        # Almacenar las instrucciones de uso como una creencia
        self.add_belief(Belief(content=instrucciones_de_uso, type=BeliefType.TOOLS))

        # Procesar cada herramienta y almacenarla como creencia
        for tool in tools:
            tool_name = tool.name
            tool_description = tool.description
            
            # Crear un texto para la herramienta
            tool_text = f"Tool Name: {tool_name}\nDescription: {tool_description}\n"
            run_method = getattr(tool, '_run', None)
            
            if run_method:
                params = inspect.signature(run_method)
                tool_text += f"Parameters: {params}\n"
            else:
                tool_text += "No _run method found.\n"
            
            tool_text += "\n---\n"
            
            # Almacenar la herramienta como una creencia
            self.add_belief(Belief(content=tool_text.strip(), type=BeliefType.TOOLS))

        
    def add_belief(self, belief: Belief):
        """Agrega o actualiza una creencia."""
        self.beliefs[belief.type][str(datetime.now())] = belief
    
    def get_beliefs_by_type(self, belief_type: BeliefType) -> List[Belief]:
        """Obtiene todas las creencias de un tipo específico."""
        return list(self.beliefs[belief_type].values())
    
    def add_desire(self, desire: Desire):
        """Agrega un nuevo deseo."""
        self.desires.append(desire)
        self.desires.sort(key=lambda x: x.priority, reverse=True)
    
    def add_intention(self, intention: Intention):
        """Agrega una nueva intención."""
        self.intentions.append(intention)
    
    def update_user_belief(self, key: str, value: str):
        """Actualiza una creencia sobre el usuario."""
        belief = Belief(
            content=f"{key}: {value}",
            type=BeliefType.USER,
            confidence=0.8,
            source="user_interaction"
        )
        self.add_belief(belief)
    
    def get_current_context(self) -> str:
        """Obtiene el contexto actual basado en las creencias."""
        context_beliefs = self.get_beliefs_by_type(BeliefType.CONTEXT)
        return "\n".join([belief.content for belief in context_beliefs])
    
    def generate_prompt_context(self) -> str:
        """Genera el contexto para el prompt, incluyendo creencias y guías sobre cómo el agente debe utilizar esta información."""
        context_parts = []

        # Instrucciones base para el agente BDI
        instructions = [
            '''''',
        ]
        context_parts.extend(instructions)

        # Agregar información de creencias núcleo (Core Beliefs)
        core_beliefs = self.get_beliefs_by_type(BeliefType.CORE)
        context_parts.append("\nCore Beliefs:")
        context_parts.extend([f"- {belief.content}" for belief in core_beliefs])

        # Agregar información de personalidad (Personality Traits)
        personality_beliefs = self.get_beliefs_by_type(BeliefType.PERSONALITY)
        if personality_beliefs:
            # Agregar una sección de escala de intensidad y los rasgos de personalidad al contexto
            context_parts.append("\nPersonality Traits:\n")
            # Agregar los rasgos de personalidad de la lista
            context_parts.extend([f"- {belief.content}" for belief in personality_beliefs])

        # Agregar información del usuario (User Information)
        user_beliefs = self.get_beliefs_by_type(BeliefType.USER)
        if user_beliefs:
            context_parts.append("\nUser Information:")
            context_parts.extend([f"- {belief.content}" for belief in user_beliefs])

        user_beliefs = self.get_beliefs_by_type(BeliefType.TOOLS)
        if user_beliefs:
            context_parts.append("\nTools Information:")
            context_parts.extend([f"- {belief.content}" for belief in user_beliefs])

        # Generación final del prompt
        return "\n".join(context_parts)


from setuptools import setup, find_packages

setup(
    name="sonika-langchain-bot",
    version="0.0.9", 
    description="Agente langchain con LLM",
    author="Erley Blanco Carvajal",
    license="MIT License",
    long_description=open("README.md", encoding="utf-8").read(),
    long_description_content_type="text/markdown",
    packages=find_packages(where="src"),  # Encuentra los paquetes dentro de "src"
    package_dir={"": "src"},  # Indica que los paquetes están en el directorio "src"
    install_requires=[
        "langchain>=0.3.0,<1.0.0",  # Permite mayor flexibilidad
        "langchain-community>=0.3.0,<1.0.0",
        "langchain-core>=0.3.5,<1.0.0",
        "langchain-openai>=0.2.0,<1.0.0",
        "langgraph>=0.2.39,<1.0.0",
        "langgraph-checkpoint>=2.0.2,<3.0.0",
        "langgraph-sdk>=0.1.34,<2.0.0",
        "dataclasses-json>=0.6.7,<1.0.0",
        "python-dateutil>=2.9.0,<3.0.0",
        "tiktoken>=0.7.0,<1.0.0",
        "pydantic>=2.9.2,<3.0.0",
        "faiss-cpu>=1.8.0,<2.0.0",
        "pypdf>=5.0.0,<6.0.0",
        "python-dotenv>=1.0.1,<2.0.0",
        "typing_extensions>=4.12.0,<5.0.0",
        "typing-inspect>=0.9.0,<1.0.0",
    ],
    extras_require={
        "dev": [
            "sphinx>=8.1.3,<9.0.0",
            "sphinx-rtd-theme>=3.0.1,<4.0.0",
        ],
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.6",  # Verifica la versión mínima de Python compatible
)
